use std::ops::{Deref, DerefMut};
use std::sync::atomic::{AtomicBool, AtomicU32, AtomicU8, Ordering};
use std::sync::{Arc, Weak};

use atomig::Atomic;
use parking_lot::lock_api::RwLockReadGuard;
use parking_lot::{RawRwLock, RwLock};

use crate::defines::{LevelInt, MAX_MOVES};
use crate::dynamic_data::event_hooks::Event;
use crate::dynamic_data::models::battle::Battle;
use crate::dynamic_data::models::learned_move::{LearnedMove, MoveLearnMethod};
use crate::dynamic_data::script_handling::{ScriptSource, ScriptSourceData, ScriptWrapper};
use crate::dynamic_data::{DynamicLibrary, Script, ScriptCategory, ScriptContainer, ScriptSet, VolatileScriptsOwner};
use crate::static_data::AbilityIndex;
use crate::static_data::Form;
use crate::static_data::Gender;
use crate::static_data::Item;
use crate::static_data::Nature;
use crate::static_data::Species;
use crate::static_data::TypeIdentifier;
use crate::static_data::{Ability, Statistic};
use crate::static_data::{ClampedStatisticSet, StatisticSet};
use crate::utils::Random;
use crate::{script_hook, PkmnResult, StringKey, ValueIdentifiable, ValueIdentifier};

/// An individual Pokemon as we know and love them.
#[derive(Debug)]
pub struct Pokemon {
    /// A unique identifier so we know what value this is.
    identifier: ValueIdentifier,
    /// The library data of the Pokemon.
    library: Arc<dyn DynamicLibrary>,
    /// The species of the Pokemon.
    species: RwLock<Arc<dyn Species>>,
    /// The form of the Pokemon.
    form: RwLock<Arc<dyn Form>>,

    /// An optional display species of the Pokemon. If this is set, the client should display this
    /// species. An example of usage for this is the Illusion ability.
    display_species: Option<Arc<dyn Species>>,
    /// An optional display form of the Pokemon. If this is set, the client should display this
    //  species. An example of usage for this is the Illusion ability.
    display_form: Option<Arc<dyn Form>>,

    /// The current level of the Pokemon.
    level: Atomic<LevelInt>,
    /// The amount of experience of the Pokemon.
    experience: AtomicU32,
    /// A unique random number for this Pokemon.
    unique_identifier: u32,

    /// The gender of the Pokemon.
    gender: RwLock<Gender>,
    /// The coloring of the Pokemon. Value 0 is the default, value 1 means shiny. Other values are
    /// currently not used, and can be used for other implementations.
    coloring: u8,
    /// The held item of the Pokemon.
    held_item: RwLock<Option<Arc<dyn Item>>>,
    /// The remaining health points of the Pokemon.
    current_health: AtomicU32,

    /// The weight of the Pokemon in kilograms.
    weight: Atomic<f32>,
    /// The height of the Pokemon in meters.
    height: Atomic<f32>,

    /// The stats of the Pokemon when disregarding any stat boosts.
    flat_stats: StatisticSet<u32>,
    /// The statistics boosts of the Pokemon. Will prevent the value from going above 6, and below
    /// -6.
    stat_boost: ClampedStatisticSet<i8, -6, 6>,
    /// The stats of the Pokemon including the stat boosts
    boosted_stats: StatisticSet<u32>,
    /// The [individual values](https://bulbapedia.bulbagarden.net/wiki/Individual_values) of the Pokemon.
    individual_values: ClampedStatisticSet<u8, 0, 31>,
    /// The [effort values](https://bulbapedia.bulbagarden.net/wiki/Effort_values) of the Pokemon.
    effort_values: ClampedStatisticSet<u8, 0, 252>,
    /// The [nature](https://bulbapedia.bulbagarden.net/wiki/Nature) of the Pokemon.
    nature: Arc<dyn Nature>,

    /// An optional nickname of the Pokemon.
    nickname: Option<String>,

    /// An index of the ability to find the actual ability on the form.
    ability_index: AbilityIndex,
    /// An ability can be overriden to an arbitrary ability. This is for example used for the Mummy
    /// ability.
    override_ability: Option<Arc<dyn Ability>>,

    /// If in battle, we have additional data.
    battle_data: RwLock<Option<PokemonBattleData>>,

    /// The moves the Pokemon has learned. This is of a set length of [`MAX_MOVES`]. Empty move slots
    /// are defined by None.
    moves: RwLock<[Option<Arc<LearnedMove>>; MAX_MOVES]>,
    /// Whether or not the Pokemon is allowed to gain experience.
    allowed_experience: bool,

    /// The current types of the Pokemon.
    types: RwLock<Vec<TypeIdentifier>>,
    /// Whether or not this Pokemon is an egg.
    is_egg: bool,
    /// Whether or not this Pokemon was caught this battle.
    is_caught: bool,

    /// The script for the held item.
    held_item_trigger_script: ScriptContainer,
    /// The script for the ability.
    ability_script: ScriptContainer,
    /// The script for the status.
    status_script: ScriptContainer,
    /// The volatile status scripts of the Pokemon.
    volatile: Arc<ScriptSet>,

    /// Data required for the Pokemon to be a script source.
    script_source_data: RwLock<ScriptSourceData>,
}

impl Pokemon {
    /// Instantiates a new Pokemon.
    pub fn new(
        library: Arc<dyn DynamicLibrary>,
        species: Arc<dyn Species>,
        form: &Arc<dyn Form>,
        ability: AbilityIndex,
        level: LevelInt,
        unique_identifier: u32,
        gender: Gender,
        coloring: u8,
        nature: &StringKey,
    ) -> Self {
        // Calculate experience from the level for the specified growth rate.
        let experience = library
            .static_data()
            .growth_rates()
            .calculate_experience(species.growth_rate(), level);
        let weight = form.weight();
        let height = form.height();
        let nature = library
            .static_data()
            .natures()
            .get_nature(nature)
            .unwrap_or_else(|| panic!("Unknown nature name was given: {}.", &nature));
        let mut pokemon = Self {
            identifier: Default::default(),
            library,
            species: RwLock::new(species),
            form: RwLock::new(form.clone()),
            display_species: None,
            display_form: None,
            level: Atomic::new(level),
            experience: AtomicU32::new(experience),
            unique_identifier,
            gender: RwLock::new(gender),
            coloring,
            held_item: RwLock::new(None),
            current_health: AtomicU32::new(1),
            weight: Atomic::new(weight),
            height: Atomic::new(height),
            stat_boost: Default::default(),
            flat_stats: Default::default(),
            boosted_stats: Default::default(),
            individual_values: Default::default(),
            effort_values: Default::default(),
            nature,
            nickname: None,
            ability_index: ability,
            override_ability: None,
            battle_data: RwLock::new(None),
            moves: RwLock::new([None, None, None, None]),
            allowed_experience: false,
            types: RwLock::new(form.types().to_vec()),
            is_egg: false,
            is_caught: false,
            held_item_trigger_script: ScriptContainer::default(),
            ability_script: ScriptContainer::default(),
            status_script: ScriptContainer::default(),
            volatile: Default::default(),
            script_source_data: Default::default(),
        };
        pokemon.recalculate_flat_stats();
        let health = pokemon.flat_stats().hp();
        pokemon.current_health = AtomicU32::new(health);

        pokemon
    }

    /// The library data of the Pokemon.
    pub fn library(&self) -> &Arc<dyn DynamicLibrary> {
        &self.library
    }
    /// The species of the Pokemon.
    pub fn species(&self) -> Arc<dyn Species> {
        self.species.read().clone()
    }
    /// The form of the Pokemon.
    pub fn form(&self) -> Arc<dyn Form> {
        self.form.read().clone()
    }
    /// The species that should be displayed to the user. This handles stuff like the Illusion ability.
    pub fn display_species(&self) -> Arc<dyn Species> {
        if let Some(v) = &self.display_species {
            v.clone()
        } else {
            self.species()
        }
    }
    /// The form that should be displayed to the user. This handles stuff like the Illusion ability.
    pub fn display_form(&self) -> Arc<dyn Form> {
        if let Some(v) = &self.display_form {
            v.clone()
        } else {
            self.form()
        }
    }
    /// The current level of the Pokemon.
    pub fn level(&self) -> LevelInt {
        self.level.load(Ordering::Relaxed)
    }
    /// The amount of experience of the Pokemon.
    pub fn experience(&self) -> u32 {
        self.experience.load(Ordering::Relaxed)
    }
    /// A unique random number for this Pokemon.
    pub fn unique_identifier(&self) -> u32 {
        self.unique_identifier
    }
    /// The gender of the Pokemon.
    pub fn gender(&self) -> Gender {
        *self.gender.read()
    }
    /// The coloring of the Pokemon. Value 0 is the default, value 1 means shiny. Other values are
    /// currently not used, and can be used for other implementations.
    pub fn coloring(&self) -> u8 {
        self.coloring
    }
    /// Gets the held item of a Pokemon
    pub fn held_item(&self) -> &RwLock<Option<Arc<dyn Item>>> {
        &self.held_item
    }
    /// Checks whether the Pokemon is holding a specific item.
    pub fn has_held_item(&self, name: &StringKey) -> bool {
        // Only true if we have an item, and the item name is the same as the requested item.
        if let Some(v) = self.held_item.read().deref() {
            return v.name() == name;
        }
        false
    }
    /// Changes the held item of the Pokemon. Returns the previously held item.
    pub fn set_held_item(&self, item: &Arc<dyn Item>) -> Option<Arc<dyn Item>> {
        self.held_item.write().replace(item.clone())
    }
    /// Removes the held item from the Pokemon. Returns the previously held item.
    pub fn remove_held_item(&self) -> Option<Arc<dyn Item>> {
        self.held_item.write().take()
    }
    /// Makes the Pokemon uses its held item.
    pub fn consume_held_item(&self) -> bool {
        if self.held_item.read().is_none() {
            return false;
        }
        let script = self
            .library
            .load_item_script(self.held_item.read().as_ref().unwrap())
            .unwrap();
        if script.is_none() {
            return false;
        }

        // TODO: the entire item use part.
        todo!();
    }

    /// The remaining health points of the Pokemon.
    pub fn current_health(&self) -> u32 {
        self.current_health.load(Ordering::Relaxed)
    }
    /// The max health points of the Pokemon.
    pub fn max_health(&self) -> u32 {
        self.boosted_stats.hp()
    }
    /// The weight of the Pokemon in kilograms.
    pub fn weight(&self) -> f32 {
        self.weight.load(Ordering::Relaxed)
    }
    /// Sets the weight of the Pokemon in kilograms.
    pub fn set_weight(&self, weight: f32) {
        self.weight.store(weight, Ordering::Relaxed)
    }

    /// The height of the Pokemon in meters.
    pub fn height(&self) -> f32 {
        self.height.load(Ordering::Relaxed)
    }
    /// An optional nickname of the Pokemon.
    pub fn nickname(&self) -> &Option<String> {
        &self.nickname
    }
    /// An index of the ability to find the actual ability on the form.
    pub fn real_ability(&self) -> &AbilityIndex {
        &self.ability_index
    }
    /// The current types of the Pokemon.
    pub fn types(&self) -> RwLockReadGuard<'_, RawRwLock, Vec<TypeIdentifier>> {
        self.types.read()
    }
    /// The moves the Pokemon has learned. This is of a set length of [`MAX_MOVES`]. Empty move slots
    /// are defined by None.
    pub fn learned_moves(&self) -> &RwLock<[Option<Arc<LearnedMove>>; MAX_MOVES]> {
        &self.moves
    }

    /// The stats of the Pokemon when disregarding any stat boosts.
    pub fn flat_stats(&self) -> &StatisticSet<u32> {
        &self.flat_stats
    }

    /// The amount of boosts on a specific stat.
    pub fn stat_boosts(&self) -> &ClampedStatisticSet<i8, -6, 6> {
        &self.stat_boost
    }

    /// The stats of the Pokemon including the stat boosts
    pub fn boosted_stats(&self) -> &StatisticSet<u32> {
        &self.boosted_stats
    }
    /// Get the stat boosts for a specific stat.
    pub fn stat_boost(&self, stat: Statistic) -> i8 {
        self.stat_boost.get_stat(stat)
    }
    /// Change a boosted stat by a certain amount.
    pub fn change_stat_boost(&self, stat: Statistic, mut diff_amount: i8, self_inflicted: bool) -> bool {
        let mut prevent = false;
        script_hook!(
            prevent_stat_boost_change,
            self,
            self,
            stat,
            diff_amount,
            self_inflicted,
            &mut prevent
        );
        if prevent {
            return false;
        }
        script_hook!(
            change_stat_boost_change,
            self,
            self,
            stat,
            self_inflicted,
            &mut diff_amount
        );
        if diff_amount == 0 {
            return false;
        }

        let mut changed = false;
        let old_value = self.stat_boost.get_stat(stat);
        match diff_amount.cmp(&0_i8) {
            std::cmp::Ordering::Less => {
                changed = self.stat_boost.decrease_stat(stat, -diff_amount);
            }
            std::cmp::Ordering::Greater => {
                changed = self.stat_boost.increase_stat(stat, -diff_amount);
            }
            _ => {}
        }
        if changed {
            if let Some(battle) = self.get_battle() {
                let new_value = self.stat_boost(stat);
                battle.event_hook().trigger(Event::StatBoostChange {
                    user: self,
                    stat,
                    old_value,
                    new_value,
                })
            }
            self.recalculate_boosted_stats();
        }
        changed
    }

    /// The [individual values](https://bulbapedia.bulbagarden.net/wiki/Individual_values) of the Pokemon.
    pub fn individual_values(&self) -> &ClampedStatisticSet<u8, 0, 31> {
        &self.individual_values
    }
    /// The [effort values](https://bulbapedia.bulbagarden.net/wiki/Effort_values) of the Pokemon.
    pub fn effort_values(&self) -> &ClampedStatisticSet<u8, 0, 252> {
        &self.effort_values
    }

    /// Gets the battle the battle is currently in.
    pub fn get_battle(&self) -> Option<&Battle> {
        let r = self.battle_data.read();
        if let Some(data) = &r.deref() {
            unsafe { data.battle.as_ref() }
        } else {
            None
        }
    }
    /// Get the index of the side of the battle the Pokemon is in. Only returns a value if the Pokemon
    /// is on the battlefield.
    pub fn get_battle_side_index(&self) -> Option<u8> {
        self.battle_data.read().as_ref().map(|data| data.battle_side_index())
    }
    /// Get the index of the slot on the side of the battle the Pokemon is in. Only returns a value
    /// if the Pokemon is on the battlefield.
    pub fn get_battle_index(&self) -> Option<u8> {
        self.battle_data.read().as_ref().map(|data| data.index())
    }
    /// Returns whether something overrides the ability.
    pub fn is_ability_overriden(&self) -> bool {
        self.override_ability.is_some()
    }
    /// Returns the currently active ability.
    pub fn active_ability(&self) -> Arc<dyn Ability> {
        if let Some(v) = &self.override_ability {
            return v.clone();
        }
        self.library
            .static_data()
            .abilities()
            .get(self.form().get_ability(self.ability_index))
            .unwrap()
    }

    /// The script for the status.
    pub fn status(&self) -> &ScriptContainer {
        &self.status_script
    }

    /// Returns the script for the currently active ability.
    pub fn ability_script(&self) -> &ScriptContainer {
        &self.ability_script
    }

    /// Whether or not the Pokemon is allowed to gain experience.
    pub fn allowed_experience_gain(&self) -> bool {
        self.allowed_experience
    }

    /// The [nature](https://bulbapedia.bulbagarden.net/wiki/Nature) of the Pokemon.
    pub fn nature(&self) -> &Arc<dyn Nature> {
        &self.nature
    }

    /// Calculates the flat stats on the Pokemon. This should be called when for example the base
    /// stats, level, nature, IV, or EV changes. This has a side effect of recalculating the boosted
    /// stats, as those depend on the flat stats.
    pub fn recalculate_flat_stats(&self) {
        self.library
            .stat_calculator()
            .calculate_flat_stats(self, &self.flat_stats);
        self.recalculate_boosted_stats();
    }
    /// Calculates the boosted stats on the Pokemon, _without_ recalculating the flat stats.
    /// This should be called when a stat boost changes.
    pub fn recalculate_boosted_stats(&self) {
        self.library
            .stat_calculator()
            .calculate_boosted_stats(self, &self.boosted_stats);
    }

    /// Change the species of the Pokemon.
    pub fn change_species(&self, species: Arc<dyn Species>, form: Arc<dyn Form>) {
        *self.species.write() = species.clone();
        *self.form.write() = form.clone();

        // If the pokemon is genderless, but it's new species is not, we want to set its gender
        if self.gender() != Gender::Genderless && species.gender_rate() < 0.0 {
            // If we're in battle, use the battle random for predictability
            let r = self.battle_data.read();
            if let Some(data) = r.deref() {
                let mut random = data.battle().unwrap().random().get_rng().lock().unwrap();
                *self.gender.write() = species.get_random_gender(random.deref_mut());
            } else {
                // If we're not in battle, just use a new random.
                *self.gender.write() = species.get_random_gender(&mut Random::default());
            }
        }
        // Else if the new species is genderless, but the pokemon has a gender, make the creature genderless.
        else if species.gender_rate() < 0.0 && self.gender() != Gender::Genderless {
            *self.gender.write() = Gender::Genderless;
        }
        let r = self.battle_data.read();
        if let Some(battle_data) = &r.deref() {
            if let Some(battle) = battle_data.battle() {
                battle.event_hook().trigger(Event::SpeciesChange {
                    pokemon: self,
                    species,
                    form,
                })
            }
        }
    }

    /// Change the form of the Pokemon.
    pub fn change_form(&self, form: &Arc<dyn Form>) {
        if self.form().value_identifier() == form.value_identifier() {
            return;
        }
        *self.form.write() = form.clone();

        {
            let mut type_lock = self.types.write();
            type_lock.clear();
            for t in form.types() {
                type_lock.push(*t);
            }
        }
        self.weight.store(form.weight(), Ordering::SeqCst);
        self.height.store(form.height(), Ordering::SeqCst);

        let ability_script = self
            .library
            .load_script(self.into(), ScriptCategory::Ability, self.active_ability().name())
            .unwrap();
        if let Some(ability_script) = ability_script {
            self.ability_script
                .set(ability_script)
                .as_ref()
                // Ensure the ability script gets initialized with the parameters for the ability.
                .on_initialize(&self.library, self.active_ability().parameters().to_vec())
        } else {
            self.ability_script.clear();
        }
        let old_health = self.max_health();
        self.recalculate_flat_stats();
        let diff_health = (self.max_health() - old_health) as i32;
        if self.current_health() == 0 && (self.current_health() as i32) < -diff_health {
            self.current_health.store(0, Ordering::SeqCst);
        } else if diff_health < 0 {
            self.current_health.fetch_sub(-diff_health as u32, Ordering::SeqCst);
        } else {
            self.current_health.fetch_add(diff_health as u32, Ordering::SeqCst);
        }
        // TODO: consider form specific attacks?

        let r = self.battle_data.read();
        if let Some(battle_data) = r.deref() {
            if let Some(battle) = battle_data.battle() {
                battle.event_hook().trigger(Event::FormChange {
                    pokemon: self,
                    form: form.clone(),
                })
            }
        }
    }

    /// Whether or not the Pokemon is useable in a battle.
    pub fn is_usable(&self) -> bool {
        !self.is_caught && !self.is_egg && !self.is_fainted()
    }

    /// Returns whether the Pokemon is fainted.
    pub fn is_fainted(&self) -> bool {
        self.current_health() == 0
    }

    /// Sets the current battle the Pokemon is in.
    pub fn set_battle_data(&self, battle: *mut Battle, battle_side_index: u8) {
        let mut w = self.battle_data.write();
        if let Some(battle_data) = w.deref_mut() {
            battle_data.battle = battle;
            battle_data.battle_side_index.store(battle_side_index, Ordering::SeqCst);
        } else {
            w.replace(PokemonBattleData {
                battle,
                battle_side_index: AtomicU8::new(battle_side_index),
                index: AtomicU8::new(0),
                on_battle_field: AtomicBool::new(false),
                seen_opponents: Default::default(),
            });
        }
    }

    /// Sets whether or not the Pokemon is on the battlefield.
    pub fn set_on_battlefield(&self, value: bool) {
        let r = self.battle_data.read();
        if let Some(data) = &mut r.deref() {
            data.on_battle_field.store(value, Ordering::SeqCst);
            if !value {
                self.volatile.clear();
                self.weight.store(self.form().weight(), Ordering::SeqCst);
                self.height.store(self.form().height(), Ordering::SeqCst);
            }
        }
    }

    /// Sets the index of the slot of the side the Pokemon is on.
    pub fn set_battle_index(&self, index: u8) {
        let r = self.battle_data.read();
        if let Some(data) = r.deref() {
            data.index.store(index, Ordering::SeqCst)
        }
    }

    /// Whether or not the Pokemon is on the battlefield.
    pub fn is_on_battlefield(&self) -> bool {
        self.battle_data.read().as_ref().is_some_and(|a| a.on_battle_field())
    }

    /// Marks an opponent as seen, for use in experience gain.
    pub fn mark_opponent_as_seen(&self, pokemon: Weak<Pokemon>) {
        let r = self.battle_data.read();
        if let Some(battle_data) = &r.deref() {
            let mut opponents = battle_data.seen_opponents().write();
            for seen_opponent in opponents.deref() {
                if seen_opponent.ptr_eq(&pokemon) {
                    return;
                }
            }
            opponents.push(pokemon);
        }
    }

    /// Damages the Pokemon by a certain amount of damage, from a damage source.
    pub fn damage(&self, mut damage: u32, source: DamageSource) {
        if damage > self.current_health() {
            damage = self.current_health();
        }
        if damage == 0 {
            return;
        }
        let new_health = self.current_health() - damage;
        if let Some(battle_data) = &self.battle_data.read().deref() {
            if let Some(battle) = battle_data.battle() {
                battle.event_hook().trigger(Event::Damage {
                    pokemon: self,
                    source,
                    original_health: self.current_health(),
                    new_health,
                });
            }
        }
        if self.battle_data.read().as_ref().is_some_and(|a| a.on_battle_field()) {
            script_hook!(on_damage, self, self, source, self.current_health(), new_health);
        }

        self.current_health.store(new_health, Ordering::SeqCst);
        if self.is_fainted() && damage > 0 {
            self.on_faint(source);
        }
    }

    /// Triggers when the Pokemon faints.
    fn on_faint(&self, source: DamageSource) {
        let r = self.battle_data.read();
        if let Some(battle_data) = r.deref() {
            if let Some(battle) = battle_data.battle() {
                battle.event_hook().trigger(Event::Faint { pokemon: self });
                script_hook!(on_faint, self, self, source);
                script_hook!(on_remove, self,);

                if !battle.can_slot_be_filled(battle_data.battle_side_index(), battle_data.index()) {
                    battle.sides()[battle_data.battle_side_index() as usize]
                        .mark_slot_as_unfillable(battle_data.index());
                }

                battle.validate_battle_state();
            }
        }
    }

    /// Heals the Pokemon by a specific amount. Unless allow_revive is set to true, this will not
    /// heal if the Pokemon has 0 health. If the amount healed is 0, this will return false.
    pub fn heal(&self, mut amount: u32, allow_revive: bool) -> bool {
        if self.current_health() == 0 && !allow_revive {
            return false;
        }
        let max_amount = self.max_health() - self.current_health();
        if amount > max_amount {
            amount = max_amount;
        }
        if amount == 0 {
            return false;
        }
        let new_health = self.current_health() + max_amount;
        if let Some(battle_data) = &self.battle_data.read().deref() {
            if let Some(battle) = battle_data.battle() {
                battle.event_hook().trigger(Event::Heal {
                    pokemon: self,
                    original_health: self.current_health(),
                    new_health,
                });
            }
        }
        self.current_health.store(new_health, Ordering::SeqCst);
        true
    }

    /// Learn a move.
    pub fn learn_move(&self, move_name: &StringKey, learn_method: MoveLearnMethod) {
        let mut learned_moves = self.learned_moves().write();
        let move_pos = learned_moves.iter().position(|a| a.is_none());
        if move_pos.is_none() {
            panic!("No more moves with an empty space found.");
        }
        let move_data = self.library.static_data().moves().get(move_name).unwrap();
        learned_moves[move_pos.unwrap()] = Some(Arc::new(LearnedMove::new(move_data, learn_method)));
    }

    /// Removes the current non-volatile status from the Pokemon.
    pub fn clear_status(&self) {
        self.status_script.clear()
    }

    /// Increases the level by a certain amount
    pub fn change_level_by(&self, amount: LevelInt) {
        self.level
            .fetch_update(Ordering::SeqCst, Ordering::Relaxed, |x| {
                let max_level = self.library().static_data().settings().maximum_level();
                if x + amount > max_level {
                    Some(max_level)
                } else {
                    Some(x + amount)
                }
            })
            .expect("Failed to change level.");
        self.recalculate_flat_stats();
    }
}

/// The data of the Pokemon related to being in a battle.
#[derive(Debug)]
pub struct PokemonBattleData {
    /// The battle data of the Pokemon
    battle: *mut Battle,
    /// The index of the side of the Pokemon
    battle_side_index: AtomicU8,
    /// The index of the slot on the side of the Pokemon.
    index: AtomicU8,
    /// Whether or not the Pokemon is on the battlefield.
    on_battle_field: AtomicBool,
    /// A list of opponents the Pokemon has seen this battle.
    seen_opponents: RwLock<Vec<Weak<Pokemon>>>,
}

impl PokemonBattleData {
    /// The battle data of the Pokemon
    pub fn battle_mut(&mut self) -> Option<&mut Battle> {
        unsafe { self.battle.as_mut() }
    }
    /// The battle data of the Pokemon
    pub fn battle(&self) -> Option<&Battle> {
        unsafe { self.battle.as_ref() }
    }

    /// The index of the side of the Pokemon
    pub fn battle_side_index(&self) -> u8 {
        self.battle_side_index.load(Ordering::Relaxed)
    }
    /// The index of the slot on the side of the Pokemon.
    pub fn index(&self) -> u8 {
        self.index.load(Ordering::Relaxed)
    }
    /// Whether or not the Pokemon is on the battlefield.
    pub fn on_battle_field(&self) -> bool {
        self.on_battle_field.load(Ordering::Relaxed)
    }
    /// A list of opponents the Pokemon has seen this battle.
    pub fn seen_opponents(&self) -> &RwLock<Vec<Weak<Pokemon>>> {
        &self.seen_opponents
    }
}

impl ScriptSource for Pokemon {
    fn get_script_count(&self) -> usize {
        let mut c = 3;
        if let Some(battle_data) = &self.battle_data.read().deref() {
            if let Some(battle) = battle_data.battle() {
                c += battle.sides()[battle_data.battle_side_index() as usize].get_script_count();
            }
        }
        c
    }

    fn get_script_source_data(&self) -> &RwLock<ScriptSourceData> {
        &self.script_source_data
    }

    fn get_own_scripts(&self, scripts: &mut Vec<ScriptWrapper>) {
        scripts.push((&self.held_item_trigger_script).into());
        scripts.push((&self.ability_script).into());
        scripts.push((&self.status_script).into());
        scripts.push((&self.volatile).into());
    }

    fn collect_scripts(&self, scripts: &mut Vec<ScriptWrapper>) {
        self.get_own_scripts(scripts);
        if let Some(battle_data) = &self.battle_data.read().deref() {
            if let Some(battle) = battle_data.battle() {
                battle.sides()[battle_data.battle_side_index() as usize].collect_scripts(scripts);
            }
        }
    }
}

impl VolatileScriptsOwner for Pokemon {
    fn volatile_scripts(&self) -> &Arc<ScriptSet> {
        &self.volatile
    }

    fn load_volatile_script(&self, key: &StringKey) -> PkmnResult<Option<Arc<dyn Script>>> {
        self.library.load_script(self.into(), ScriptCategory::Pokemon, key)
    }
}

impl ValueIdentifiable for Pokemon {
    fn value_identifier(&self) -> ValueIdentifier {
        self.identifier
    }
}

/// A source of damage. This should be as unique as possible.
#[derive(Debug, Clone, Copy)]
#[repr(u8)]
pub enum DamageSource {
    /// The damage is done by a move.
    MoveDamage = 0,
    /// The damage is done by something else.
    Misc = 1,
    /// The damage is done because of struggling.
    Struggle = 2,
}

#[cfg(test)]
pub mod test {
    use crate::dynamic_data::libraries::test::MockDynamicLibrary;
    use crate::dynamic_data::models::pokemon::Pokemon;
    use crate::dynamic_data::tests::MockBattleStatCalculator;
    use crate::dynamic_data::DynamicLibrary;
    use crate::static_data::growth_rate_library::tests::MockGrowthRateLibrary;
    use crate::static_data::nature_library::tests::MockNatureLibrary;
    use crate::static_data::species_library::tests::MockSpeciesLibrary;
    use crate::static_data::test::MockStaticData;
    use crate::static_data::tests::{MockForm, MockNature, MockSpecies};
    use crate::static_data::{AbilityIndex, Species};
    use crate::static_data::{Form, Gender};
    use crate::StringKey;
    use std::sync::Arc;

    pub fn mock_library() -> Arc<dyn DynamicLibrary> {
        let mut species_lib = MockSpeciesLibrary::new();
        species_lib.expect_get().return_once(|_| {
            let mut species = MockSpecies::new();
            species.expect_get_form().return_once(|_| {
                let mut form = MockForm::new();
                form.expect_name().return_const(StringKey::new("default"));
                form.expect_weight().return_const(100.0);
                form.expect_height().return_const(20.0);
                form.expect_types().return_const(vec![0u8.into()]);
                let a: Arc<dyn Form> = Arc::new(form);
                Some(a)
            });
            species.expect_growth_rate().return_const(StringKey::empty());
            species.expect_name().return_const(StringKey::new("test_species"));

            let s: Arc<dyn Species> = Arc::new(species);
            Some(s)
        });

        let mut static_lib = MockStaticData::new();
        static_lib.expect_species().return_const(Box::new(species_lib));

        let mut growth_rate_lib = MockGrowthRateLibrary::new();
        growth_rate_lib.expect_calculate_experience().return_const(1000u32);

        let mut nature_lib = MockNatureLibrary::new();
        nature_lib.expect_get_nature().returning(|_| {
            let n = MockNature::new();
            Some(Arc::new(n))
        });

        static_lib.expect_growth_rates().return_const(Box::new(growth_rate_lib));
        static_lib.expect_natures().return_const(Box::new(nature_lib));

        let mut stat_calculator = MockBattleStatCalculator::new();
        stat_calculator.expect_calculate_flat_stats().returning(|_, _| {});
        stat_calculator.expect_calculate_boosted_stats().returning(|_, _| {});

        let mut lib = MockDynamicLibrary::new();
        lib.expect_static_data().return_const(Box::new(static_lib));
        lib.expect_stat_calculator().return_const(Box::new(stat_calculator));

        Arc::new(lib)
    }

    #[test]
    fn construct_pokemon() {
        let lib = mock_library();
        let species = lib.static_data().species().get(&"foo".into()).unwrap();
        let form = species.get_form(&"default".into()).unwrap();

        let pokemon = Pokemon::new(
            lib,
            species,
            &form,
            AbilityIndex {
                hidden: false,
                index: 0,
            },
            10,
            0,
            Gender::Male,
            0,
            &"test_nature".into(),
        );
        assert_eq!(pokemon.species().name(), &"test_species".into());
        assert_eq!(pokemon.form().name(), &"default".into());
    }
}
