use crate::ffi::{ffi_arc_dyn_getter, ExternPointer, IdentifiablePointer, OwnedPtr};
use crate::static_data::{BattleItemCategory, Item, ItemCategory, ItemImpl};
use crate::StringKey;
use hashbrown::HashSet;
use std::ffi::{c_char, CStr, CString};
use std::ptr::drop_in_place;
use std::sync::Arc;

/// Instantiates an item.
#[no_mangle]
unsafe extern "C" fn item_new(
    name: *const c_char,
    category: ItemCategory,
    battle_category: BattleItemCategory,
    price: i32,
    flags: *const *const c_char,
    flags_length: usize,
) -> IdentifiablePointer<Arc<dyn Item>> {
    let flags = std::slice::from_raw_parts(flags, flags_length);
    let name: StringKey = CStr::from_ptr(name).to_str().unwrap().into();
    let mut flags_set: HashSet<StringKey> = HashSet::with_capacity(flags_length);
    for flag in flags {
        flags_set.insert(CStr::from_ptr(*flag).to_str().unwrap().into());
    }

    let item: Arc<dyn Item> = Arc::new(ItemImpl::new(&name, category, battle_category, price, flags_set));
    item.into()
}

/// Drops a reference counted item.
#[no_mangle]
unsafe extern "C" fn item_drop(ptr: OwnedPtr<Arc<dyn Item>>) {
    drop_in_place(ptr)
}

/// The name of the item.
#[no_mangle]
unsafe extern "C" fn item_name(ptr: ExternPointer<Arc<dyn Item>>) -> OwnedPtr<c_char> {
    let name = ptr.as_ref().name();
    CString::new(name.str()).unwrap().into_raw()
}

ffi_arc_dyn_getter!(Item, category, ItemCategory);
ffi_arc_dyn_getter!(Item, battle_category, BattleItemCategory);
ffi_arc_dyn_getter!(Item, price, i32);

/// Checks whether the item has a specific flag.
#[no_mangle]
unsafe extern "C" fn item_has_flag(ptr: ExternPointer<Arc<dyn Item>>, flag: *const c_char) -> u8 {
    let flag = CStr::from_ptr(flag).into();
    u8::from(ptr.as_ref().has_flag(&flag))
}
